package Config::Model::Dpkg::Copyright ;

use strict;
use warnings;

use 5.20.0;
use IO::Pipe;

use feature qw/postderef signatures/;
no warnings qw/experimental::postderef experimental::signatures/;

use base qw/Config::Model::Node/;
use Path::Tiny;

use Dpkg::Copyright::Scanner qw/scan_files/;
use Software::LicenseUtils;
use Scalar::Util qw/weaken/;

sub update ($self) {

    my @copyright_data = scan_files();

    my %old_files;
    my %old_split_files;
    # load existing path data to remove duplicates
    foreach my $f ($self->grab("Files")->fetch_all_indexes) {
        my @paths = sort split /\s+/,$f;
        $old_files{"@paths"} = [ \$f, $f ];
        foreach my $path (@paths) {
            next if $path =~ /\*/;
            $old_split_files{$path} = \$f ;
        }
        # the weakened ref will vanish if all ref to $f are removed from %old_split_files
        weaken($old_files{"@paths"}[0]);
    }

    foreach my $data (@copyright_data) {
        my ($paths, $c, $l) = $data->@*;
        #say "load '@$paths->[0]' with '$c' '$l'";

        # remove paths from old stuff that are found in current list
        delete $old_files{ join (' ', sort @$paths)} ;
        map { delete $old_split_files{$_} } @$paths;

        next if $c eq 'no-info-found';

        my $c_obj = $self->grab( qq!Files:"@$paths"!);
        $c_obj->load(qq!Copyright="$c"!);
        my $short_obj = $c_obj->grab( qq!License short_name!);

        # skip when file contains actual information and extracted
        # license is unknown
        next if $l eq 'UNKNOWN' and $short_obj->fetch();

        # handle the case where license is something like GPL-2 or GPL-3
        foreach my $sub_l (split / or /, $l) {
            my $license_object ;
            eval {
                $license_object = Software::LicenseUtils->new_from_short_name( {
                    short_name => $sub_l,
                    holder => 'X. Ample'
                }) ;
            };
            if ($license_object) {
                $self->load(qq!License:$sub_l!);
            }
        }

        $short_obj->store($l);

        if ($short_obj->has_error) {
            my $text="Please fill license $l from header of ".$paths->[0];
            $c_obj->load(qq!License full_license="$text" short_name="$l"!);
        }

    }

    # now %old_files contains entries that were not touched by the
    # found licenses.  either the files were removed, or the path
    # distributions don't match
    foreach my $old_sorted_entry (sort keys %old_files) {
        my ($weak_old_ref, $old_entry) = $old_files{$old_sorted_entry}->@*;
        next if $old_entry =~ /debian/; # leave debian dir alone
        if (not defined $weak_old_ref) {
            # all paths of $old_entry were found but arranged differently
            say "deleting old entry: '$old_entry'";
            $self->grab("Files")->delete($old_entry);
        }
        else {
            # %old_split_files contains paths that were not found in the
            # license process. i.e removed files.
            my @new_paths;
            foreach my $path (split /\s+/,$old_sorted_entry) {
                if (defined $old_split_files{$path}) {
                    push @new_paths, $path;
                }
                else {
                    say "deleting $path from entry '$old_entry'";
                }
            }
            my $new_path_entry = join(' ',@new_paths);
            if ($new_path_entry ne $old_sorted_entry) {
                say "renaming $old_entry in $new_path_entry";
                $self->grab("Files")->move($old_entry, $new_path_entry);
            }
       }
    }

    # read a debian/fix.copyright file to patch scanned data
    my $fix = path('debian')->child('fix-scanned-copyright');
    if ($fix->exists) {
        foreach my $l ( $fix->lines) {
            $self->load( $l );
        }
    }


    return ''; # improve returned message ?
}

1;

__END__


=head1 NAME

 Config::Model::Dpkg::Copyright - Fill the File sections of debian/copyright file

=head1 SYNOPSIS

 # this modules is used by cme when invoked with this command
 $ cme update dpkg-copyright

=head1 DESCRIPTION

This commands helps with the tedious task of maintening
C<debian/copyright> file. When you package a new release of a
software, you can run C<cme update dpkg-copyright> to update the
content of the copyright file.

This command scans current package directory to extract copyright and
license information and store them in the Files sections of
debian/copyright file.

In debian package directory:

* run 'cme update dpkg-copyright' or 'cme update dpkg'
* check the result with your favorite VCS diff tool. (you do use
  a VCS for your package files, do you ?)

Note: this command is experimental.

=head1 Tweak results

Since the extraction of copyright information from source file is
based on comments, the result is sometimes lackluster. Your may
specify instruction to alter or set specific copyright entries in
C<debian/fix.copyright> file. Each line of this file will be handled
by L<Config::Model::Loader> to modify copyright information.

=head2 Example

If the extracted copyright contains:

 Files: *
 Copyright: 2014-2015, Adam Kennedy <adamk@cpan.org> "foobar
 License: Artistic or GPL-1+

You may add this line in C<debian/fix.copyright> file:

 Files:'*' Copyright=~s/\s*".*//

This way, the copyright information will be updated from the file
content but the extra C<"foobar> will always be removed during
updates.

=head1 AUTHOR

Dominique Dumont <dod@debian.org>

=cut
