package Dpkg::Copyright::Scanner ;

use strict;
use warnings;

use 5.20.0;
use IO::Pipe;
use Exporter::Lite;
use Array::IntSpan;

use feature qw/postderef signatures/;
no warnings qw/experimental::postderef experimental::signatures/;

our @EXPORT = qw(scan_files print_copyright);

my $whitespace_list_delimiter = $ENV{'whitespace_list_delimiter'} || "\n ";

# license and copyright sanitisation pilfered from Jonas's licensecheck2dep5
# hence this file is GPL-2+ not LGPL-2.1+

# Copyright 2014 Dominique Dumont <dod@debian.org>
# Copyright © 2005-2012 Jonas Smedegaard <dr@jones.dk>
# Description: Reformat licencecheck output to copyright file format
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

sub print_copyright ( %args ){
    my @copyright_data = scan_files(%args);

    my @out ;

    foreach my $data (@copyright_data) {
        my ($paths, $c, $l) = $data->@*;
        next if $c eq 'no-info-found';
        push @out,
            "Files: ", join($whitespace_list_delimiter, $paths->@* )."\n",
            "Copyright: $c\n",
            "License: $l\n", "\n";
    }

    if ($args{out}) {
        $args{out}->spew_utf8( @out);
    }
    else {
        binmode(STDOUT, ":utf8");
        print @out;
    }
}

# option to skip UNKNOWN ?
# load a file to override some entries ?
sub scan_files ( %args ) {

    my @lines ;
    if ($args{in}) {
        @lines = $args{in}->lines_utf8; # for tests
    }
    else {
        my $pipe = IO::Pipe->new();
        $pipe->reader("licensecheck --copyright -m -r .");
        binmode($pipe, ":utf8");
        @lines = $pipe->getlines;
        $pipe->close;
    }

    my %copyrights ;
    my $files = {};
    my $id = 0;

    foreach my $line (sort @lines) {
        chomp $line;
        # say "found: $line";
        my ($f,$l,$c) = split /\t/, $line; 
        $f =~ s!\./!!;
        $l =~ s/([*?\\])/\\$1/g;
        $l =~ s/\s*\(unversioned\/unknown version\)//;
        $l =~ s/\s*\(with incorrect FSF address\)//;
        $l =~ s/(\w+)\s+\(v([^)]+) or v([^)]+)\)/uc($1)."-$2 or ".uc($1)."-$3"/e;
        $l =~ s/\s+\(v([^)]+) or later\)/-$1+/;
        $l =~ s/\s+\(v([^)]+)\)/-$1/;
        $l =~ s/^\s*(GENERATED FILE)/UNKNOWN/;
        $l =~ s/\s+(GENERATED FILE)//;
        $l =~ s/\bzlib\/libpng\b/Zlib/;
        $l =~ s/\bMIT\/X11 \(BSD like\)/Expat/;
        $l =~ s/^\s*BSD \((\d) clause\)$/BSD-$1-clause/;
        $l =~ s/^\s*public domain$/public-domain/i;

        # this is very fragile. may need to change license-check to output license keyword
        $l =~ s/ / and /g;

        $c =~ s/'//g;
        $c =~ s/^&copy;\s*//;
        $c =~ s/(?<=\b\d{4})\s*-\s*\d{4}(?=\s*-\s*(\d{4})\b)//g;
        $c =~ s/(\d+)\s*-\s*(\d+)/$1-$2/g;
        $c =~ s/\b(\d{4}),?\s+([\S^\d])/$1, $2/g;
        $c = 'no-info-found' if $c =~ /^\*No/;

        $c =  __pack_copyright($c);

        my $a = $copyrights{$c}{$l} //= $id++;
        # split file path and fill recursive hash, leaf is id
        my @path = split m!/!,$f;
        my $file = pop @path;
        my $tmp = $files ;
        map { $tmp = $tmp->{$_} ||= {};  } @path;
        $tmp->{$file} = $a;
    }

    my @copyrights_by_id ;
    foreach my $c (sort keys %copyrights) {
        foreach my $l (sort keys $copyrights{$c}->%* ) {
            my $id = $copyrights{$c}{$l};
            $copyrights_by_id[$id] = [ $c, $l ] ;
        }
    }

    say "No copyright information found" unless keys %$files;

    # regroup %files hash: all leaves have same id -> wild card
    my $squashed = __squash($files);

    # pack files by copyright id
    my @packed = __pack_files($files);

    my @copyright_data;

    foreach my $p (@packed) {
        my ($id, @paths) = $p->@*;
        my ($c,$l) = $copyrights_by_id[$id]->@*;
        push @copyright_data, [ \@paths, $c, $l ];
    }

    return @copyright_data;
}

sub __pack_copyright ($r) {

    return $r if $r eq 'no-info-found';
    my %cop;
    $r =~ /^[\s\W]+|[\s\W]+$/g;
    foreach my $c ( split( m!\s*/\s*!, $r)) {
        my @data = split /(?<=\d)[,\s]+/, $c;
        my $owner = pop @data;
        return $r unless defined $owner;
        $owner =~ s/^\s+|\s+$//g;
        $cop{$owner} ||= [] ;
        push $cop{$owner}->@*, @data ;
    }

    my @res ;
    foreach my $owner (sort keys %cop) {
        my $span = Array::IntSpan->new();
        my $data = $cop{$owner};
        foreach my $year ($data->@*) {
            return $r if $year =~ /[^\d-]/; # bail-out
            $span->set_range_as_string($year, $owner);
        }
        $span->consolidate();
        push @res, $span->get_range_list. ', '. $owner;
    }

    return join("\n  ",reverse sort @res);
}

#in each directory, pack files that have the same copyright/license information
# traverse recursively %h (whose strcuture matches the scanned directory)
# @path keeps track of the recursion depth to provide the file path
sub __pack_files ($h) {

    my @res ;
    __pack_dir($h,\@res) ;

    # sort by first path listed in there
    my $sort_path = sub {
        $a->[1] cmp $b->[1];
    };

    return sort $sort_path @res ;
}

sub __pack_dir ($h, $pack, @path) {
    my %pack_by_id;
    foreach my $file (sort keys %$h) {
        my $id = $h->{$file};
        if (ref($id)) {
            __pack_dir($id, $pack, @path, $file) ;
        }
        elsif (defined $pack_by_id{$id} ) {
            push $pack_by_id{$id}->@*, join('/',@path,$file);
        }
        else {
            $pack_by_id{$id} = [ join('/',@path,$file) ] ;
        }
    }

    push $pack->@*, map { [ $_, $pack_by_id{$_}->@* ];  } keys %pack_by_id ;
}

# $h is a tree of hash matching the directory structure. Each leaf is a
# copyright id.
sub __squash ($h) {
    my %count ;

    # count the number of times each (c) info is used in this directory.
    # (including the main (c) info of each subdirectory)
    foreach my $name (sort keys %$h) {
        my $item = $h->{$name};
        if (ref($item)) {
            # squash may return a plain id, or a hash with '*' => id ,
            # or a non squashable hash
            $h->{$name} = __squash($item);
        }
        my $id = (ref($item) and defined $item->{'*'}) ? $item->{'*'} : $item ;

        # do not count non squashable hashes (i.e. there's no main (c) info)
        if (not ref ($id)) {
            $count{$id}//=0;
            $count{$id} ++;
        }
    }

    # find the most used (c) info in this directory
    my $max = 0;
    my $max_id;
    foreach my $id (sort keys %count) {
        if ($count{$id} > $max) {
            $max = $count{$id};
            $max_id = $id ;
        }
    }

    # all files associated to the most used (c) info are deleted to
    # be represented by '*' entry
    foreach my $name (sort keys %$h) {
        my $item = $h->{$name};
        if (ref($item) and defined $item->{'*'} and $item->{'*'} == $max_id) {
            # delete ./item/* which is covered by ./*
            delete $item->{'*'};
            # delete ./item if no files with different (c) info are there
            delete $h->{$name} unless keys $h->{$name}->%*;
        }
        if (not ref ($item)) {
            # delete file that is represented by '*' entry
            delete $h->{$name} if $item == $max_id;
        }
    }
    # here's the '*' file representing the most used (c) info
    $h->{'*'} = $max_id if defined $max_id;

    return $h;
}

1;

__END__

=head1 NAME

 Dpkg::Copyright::Scanner - Scan files to provide copyright data

=head1 SYNOPSIS

 use Dpkg::Copyright::Scanner qw/print_copyright scan_files/;

 # print copyright data on STDOUT
 print_copyright;

 # return a data structure containing copyright information
 my @copyright_data = scan_files();


=head1 DESCRIPTION

This modules scans current package directory to extract copyright and
license information. Information are packed in a way to ease review and
maintenance. Files information is grouped with wildcards ('*') to reduce
the list of files.

=head1 METHODS

=head2 print_copyright

Print copyright information on STDOUT like L<scan-copyrights>.

=head2 scan_files

Return a data structure with copyright and license information.

The structure is a list of list:

 [
   [
     [ path1 ,path2, ...],
     copyright,
     license_short_name
   ],
   ...
 ]

Example:

 [
  [
    [ '*' ],
    '1994-2001, by Frank Pilhofer.',
    'GPL-2+'
  ],
  [
    [ 'pan/*' ],
    '2002-2006, Charles Kerr <charles@rebelbase.com>',
    'GPL-2'
  ],
  [
    [
      'pan/data/parts.cc',
      'pan/data/parts.h'
    ],
    '2002-2007, Charles Kerr <charles@rebelbase.com>',
    'GPL-2'
  ],
 ]

=head1 Encoding

The output of L<licensecheck> is expected to be utf-8. Which means
that the source files scanned by L<licensecheck> should also be
encoded in utf-8. In practice, this will impact only copyright owner
name which may be garbled if comments are not encoded in utf-8.

=head1 BUGS

Extracting license and copyright data from unstructured comments is not reliable.
User must check manually the files when no copyright info is found or when the
license is unknown.

=head1 SEE ALSO

L<license_check>, C<licensecheck2dep5> from C<cdbs> package

=head1 AUTHOR

Dominique Dumont <dod@debian.org>

=cut

